/*
 * AMTP player - tool playing sound received via IEEE 1394, Audio
 *               Music Data Transmission Protocol on the soundcard.
 *
 * Copyright (C) 2002, Christian Gemeiner,
 *                     Manfred Weihs <weihs@ict.tuwien.ac.at>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include  <unistd.h>
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <fcntl.h>
#include  <sys/types.h>
#include  <sys/stat.h>
#include  <sys/ioctl.h>
#include  <sys/time.h>
#include  <sys/soundcard.h>
#include  <libraw1394/raw1394.h>
#include  "playamdtpmodule.h"

#define DEVICE_NAME "/dev/dsp"

#define MAX_CHANNELS 6

/*#define DEBUG */

int running,bufsize;
audio_buf_info info;
unsigned int isochannel=63;

static raw1394handle_t handle;
static int port=0;
static int fd_audio=-1;
static unsigned short channels;
static unsigned int samplingsp;
static unsigned short savedchannels=0;
static unsigned int savedsamplingsp=0;
static int fd=-1;
static int packetcounter=0;
static int doublesample=0;
static int discard=0;

static void Allocate_Soundcard(void);

/*Writes the data from IEEE 1394 packets to the soundcard, controls the soundcardbuffer*/
/*and sets the samplingspeed and channelnr of the soundcard specified in the packet*/

static int iso_packets(raw1394handle_t handle, int channel, size_t length,
                             quadlet_t *data)
{ 
    int address=14;
    int i,z,dbt;
    unsigned char BetoLe[2];
    unsigned short datablocks;   
    unsigned short cip;
    unsigned char audiodata[MAX_CHANNELS*2];
    int bufbytes;
    
    
    packetcounter = (packetcounter + 1) % 2;

    channels=((unsigned char *)data)[5];
    
    if (channels > MAX_CHANNELS) {
	fprintf(stderr,"too many channels, ignoring some\n");
    }

    switch(((unsigned char*)data)[9] & 0x07)
    {
	case 0x00:
	    samplingsp=32000;
            break;
        case 0x01:
            samplingsp=44100;
            break;
        case 0x02:
            samplingsp=48000;
            break;
        case 0x03:
            samplingsp=88200;
            break;
        case 0x04:
            samplingsp=96000;
            break;
        case 0x05:
            samplingsp=176400;
            break;
        case 0x06:
            samplingsp=192000;
            break; 
        default:
            (void) fprintf(stderr,"No valid sampling speed or no Basic AM824data format");
            exit(1);
    }
    
    if ((savedchannels != channels) || (savedsamplingsp != samplingsp))
    {
        Allocate_Soundcard();
        savedchannels=channels;
        savedsamplingsp=samplingsp;    
    }
    
    if ((((unsigned char *)data)[2] & 0x40) == 0x40)
    {
	cip=8u;
    }
    else
    {
	cip=0u;
    }
    
    if (!((((unsigned char*)data)[10]==0xff) && (((unsigned char*)data)[11] == 0xff)))
    {
        
        BetoLe[0]=((unsigned char *)data)[1];
        BetoLe[1]=((unsigned char *)data)[0];
        
        datablocks=((*((unsigned short *)BetoLe))-cip)/(channels*4u);
        if (!packetcounter) {
                if (ioctl(fd_audio,SNDCTL_DSP_GETOSPACE, &info)==-1)
                {  
                   perror("SND_DSP_GETOSPACE");
                   exit(1);                           
                } 

                bufbytes=info.fragstotal*info.fragsize;
                bufsize=(info.bytes*100)/bufbytes;


            if ((bufsize>75) && !doublesample)
            {
		
                #ifdef DEBUG
                   fprintf(stderr,"buffer too empty, start doubling sample\n");
		#endif
                doublesample=1;
            }
            if ((bufsize<25) && !discard)
            {
		#ifdef DEBUG
                   fprintf(stderr,"buffer too full; start discarding samples\n");
                #endif
                discard=1;
            }   


            if ((bufsize<=50) && (doublesample))
            {
		#ifdef DEBUG
                   fprintf(stderr,"stop doubling samples \n");
                #endif
		doublesample=0;  
            }
	    if ((bufsize>=50) && (discard))
            {
		#ifdef DEBUG
                    fprintf(stderr,"stop discarding samples \n");
                #endif
                discard=0;
            } 
            
	}

        /*copies the data to the soundcard*/
        for (i=1;i<=datablocks;i++)
        {
            dbt=0;
            for (z=1;z<=channels;z++)            
            { 
		if (z <= MAX_CHANNELS)
		{
		    audiodata[dbt]= ((unsigned char*)data)[address];
		    dbt++;
		    audiodata[dbt]= ((unsigned char*)data)[address-1];
		    dbt++;
		}
               address+=4;
            }
	    
            if (!discard)
            {
		(void) write(fd_audio,audiodata,2*(channels < MAX_CHANNELS?channels:MAX_CHANNELS));
                 
                if (doublesample)                 
		{
		    (void) write(fd_audio,audiodata,2*(channels < MAX_CHANNELS?channels:MAX_CHANNELS));
		}
            }

          
       }     
    }
                
   return 0;
}

/* Allocates the IEEE1394 bus. This sets handler,port, channel and fd*/ 
static void Allocate_IEEE1394(void)
{
    handle=raw1394_new_handle();
    if (!handle)
    {
	if (!errno) perror("incompatible with kernel");     
        else perror("cannot get handle");
        exit(EXIT_FAILURE);
    }       
    do
    {
	if (raw1394_get_port_info(handle,NULL,0)<=port)
        {
	    perror("port not available");
            exit(EXIT_FAILURE);
        }
        raw1394_set_port(handle,port);
    }while (errno==ESTALE); 
    fd=raw1394_get_fd(handle);
    raw1394_set_iso_handler(handle,isochannel,iso_packets);
    raw1394_start_iso_rcv(handle,isochannel);
} 

/*Allocates the soundcard. This sets the samplingspeed, channels(stereo or audio) and the data format*/
static void Allocate_Soundcard(void)
{
    int format;
    unsigned int channel;    
    unsigned int speed;
 
    packetcounter=0;
    if (fd_audio != -1)
    {      
	(void) close(fd_audio);
        fd_audio = -1;
    }
    format= AFMT_S16_LE;   
    channel=channels;
    speed=samplingsp;    

    if ((fd_audio=open(DEVICE_NAME,O_WRONLY,0)) == -1)
    {
	perror(DEVICE_NAME);
        exit(1);
    }
    
    if (ioctl(fd_audio,SNDCTL_DSP_SETFMT, &format) == -1)
    {
        perror("SNDCTL_DSP_SETFMT");
        exit(1);
    }
    if (format != AFMT_S16_LE)
    {
	(void) fprintf(stderr,"Does not support AFMT_S32_LE");    
        exit(1);
    }  
    
    if (ioctl(fd_audio,SNDCTL_DSP_CHANNELS, &channel) == -1)
    {
	perror("SNDCTL_DSP_CHANNELS");
        exit(1);
    }
    if (channel != channels)
    {
	(void) fprintf(stderr,"Device does not support stereo mode");
        exit(1);
    }     
    
    if (ioctl(fd_audio,SNDCTL_DSP_SPEED, &speed)==-1)
    {
	perror("SNDCTL_DSP_SPEED");
        exit(1);
    }
    if (speed != samplingsp)
    {
	(void) fprintf(stderr,"Device does not support sampling rate of 44,1kHz");
        exit(1);
    }    
}

/*Deallocates the IEEE1394 bus*/
static void Deallocate(void)
{
    savedchannels=0;
    savedsamplingsp=0;

    raw1394_set_iso_handler(handle,isochannel,NULL);
    raw1394_stop_iso_rcv(handle,isochannel);
    raw1394_destroy_handle(handle);

    if (fd_audio != -1)
    {      
	(void) close(fd_audio);
        fd_audio=-1;
    }
    
    fd=-1;
    running=0;
    packetcounter=0;
}


/*Checks if data is on the bus channel and starts processing packets */
void* play(void* arg)
{
    fd_set rfds;
    struct timeval tv; 
    int haveData = 1;
    Allocate_IEEE1394(); 
   
     while(running) {

	 FD_ZERO(&rfds);
	 FD_SET(fd,&rfds);
	 tv.tv_sec=0;
	 tv.tv_usec=50000;
    
    
	 switch(select(fd+1,&rfds,NULL,NULL,&tv))
         {
	    case 0:
                if (haveData) 
                {    
                     printf("\nNo data on this channel!\n");
                     fflush(stdout);
                     haveData = 0;
                } 
                break;
            case 1:    
                if (!haveData)
                {
		    printf("data on this channel again");
                    fflush(stdout);
                    haveData = 1;
                }
                raw1394_loop_iterate(handle);
                break;
            default:
		fprintf(stderr,"An error occured during select");
                exit(1);
        }
         


    }
    Deallocate(); 
  
    return NULL;
}
